"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any

import aiohttp

from . import __version__
from .backoff import Backoff
from .enums import NodeStatus
from .exceptions import AuthorizationFailedException, NodeException
from .payloads import *
from .tracks import Playable


if TYPE_CHECKING:
    from .node import Node
    from .player import Player
    from .types.request import UpdateSessionRequest
    from .types.response import InfoResponse
    from .types.state import PlayerState
    from .types.websocket import TrackExceptionPayload, WebsocketOP


logger: logging.Logger = logging.getLogger(__name__)


class Websocket:
    def __init__(self, *, node: Node) -> None:
        self.node = node

        self.backoff: Backoff = Backoff()

        self.socket: aiohttp.ClientWebSocketResponse | None = None
        self.keep_alive_task: asyncio.Task[None] | None = None

    @property
    def headers(self) -> dict[str, str]:
        assert self.node.client is not None
        assert self.node.client.user is not None

        data = {
            "Authorization": self.node.password,
            "User-Id": str(self.node.client.user.id),
            "Client-Name": f"Wavelink/{__version__}",
        }

        if self.node.session_id:
            data["Session-Id"] = self.node.session_id

        return data

    def is_connected(self) -> bool:
        return self.socket is not None and not self.socket.closed

    async def _update_node(self) -> None:
        if self.node._resume_timeout > 0:
            udata: UpdateSessionRequest = {"resuming": True, "timeout": self.node._resume_timeout}
            await self.node._update_session(data=udata)

        info: InfoResponse = await self.node._fetch_info()
        if "spotify" in info["sourceManagers"]:
            self.node._spotify_enabled = True

    async def connect(self) -> None:
        self.node._status = NodeStatus.CONNECTING

        if self.keep_alive_task:
            try:
                self.keep_alive_task.cancel()
            except Exception as e:
                logger.debug(
                    "Failed to cancel websocket keep alive while connecting. This is most likely not a problem and will not affect websocket connection: '%s'",
                    e,
                )

        retries: int | None = self.node._retries
        session: aiohttp.ClientSession = self.node._session
        heartbeat: float = self.node.heartbeat
        uri: str = f"{self.node.uri.removesuffix('/')}/v4/websocket"
        github: str = "https://github.com/PythonistaGuild/Wavelink/issues"

        while True:
            try:
                self.socket = await session.ws_connect(url=uri, heartbeat=heartbeat, headers=self.headers)  # type: ignore
            except Exception as e:
                if isinstance(e, aiohttp.WSServerHandshakeError) and e.status == 401:
                    await self.cleanup()
                    raise AuthorizationFailedException from e
                elif isinstance(e, aiohttp.WSServerHandshakeError) and e.status == 404:
                    await self.cleanup()
                    raise NodeException from e
                else:
                    logger.warning(
                        'An unexpected error occurred while connecting %r to Lavalink: "%s"\nIf this error persists or wavelink is unable to reconnect, please see: %s',
                        self.node,
                        e,
                        github,
                    )

            if self.is_connected():
                self.keep_alive_task = asyncio.create_task(self.keep_alive())
                break

            if retries == 0:
                logger.warning(
                    '%r was unable to successfully connect/reconnect to Lavalink after "%s" connection attempt. This Node has exhausted the retry count.',
                    self.node,
                    retries + 1,
                )

                await self.cleanup()
                break

            if retries:
                retries -= 1

            delay: float = self.backoff.calculate()
            logger.info('%r retrying websocket connection in "%s" seconds.', self.node, delay)

            await asyncio.sleep(delay)

    async def keep_alive(self) -> None:
        assert self.socket is not None

        while True:
            message: aiohttp.WSMessage = await self.socket.receive()

            if message.type in (  # pyright: ignore[reportUnknownMemberType]
                aiohttp.WSMsgType.CLOSED,
                aiohttp.WSMsgType.CLOSING,
            ):
                asyncio.create_task(self.connect())
                break

            if message.data is None:  # pyright: ignore[reportUnknownMemberType]
                logger.debug("Received an empty message from Lavalink websocket. Disregarding.")
                continue

            data: WebsocketOP = message.json()

            if data["op"] == "ready":
                resumed: bool = data["resumed"]
                session_id: str = data["sessionId"]

                self.node._status = NodeStatus.CONNECTED
                self.node._session_id = session_id

                await self._update_node()

                ready_payload: NodeReadyEventPayload = NodeReadyEventPayload(
                    node=self.node, resumed=resumed, session_id=session_id
                )
                self.dispatch("node_ready", ready_payload)

            elif data["op"] == "playerUpdate":
                playerup: Player | None = self.get_player(data["guildId"])
                state: PlayerState = data["state"]

                updatepayload: PlayerUpdateEventPayload = PlayerUpdateEventPayload(player=playerup, state=state)
                self.dispatch("player_update", updatepayload)

                if playerup:
                    asyncio.create_task(playerup._update_event(updatepayload))

            elif data["op"] == "stats":
                statspayload: StatsEventPayload = StatsEventPayload(data=data)
                self.node._total_player_count = statspayload.players
                self.dispatch("stats_update", statspayload)

            elif data["op"] == "event":
                player: Player | None = self.get_player(data["guildId"])

                if data["type"] == "TrackStartEvent":
                    track: Playable = Playable(data["track"])

                    startpayload: TrackStartEventPayload = TrackStartEventPayload(player=player, track=track)
                    self.dispatch("track_start", startpayload)

                    if player:
                        asyncio.create_task(player._track_start(startpayload))

                elif data["type"] == "TrackEndEvent":
                    track: Playable = Playable(data["track"])
                    reason: str = data["reason"]

                    if player and reason != "replaced":
                        player._current = None

                    endpayload: TrackEndEventPayload = TrackEndEventPayload(player=player, track=track, reason=reason)
                    self.dispatch("track_end", endpayload)

                    if player:
                        asyncio.create_task(player._auto_play_event(endpayload))

                elif data["type"] == "TrackExceptionEvent":
                    track: Playable = Playable(data["track"])
                    exception: TrackExceptionPayload = data["exception"]

                    excpayload: TrackExceptionEventPayload = TrackExceptionEventPayload(
                        player=player, track=track, exception=exception
                    )
                    self.dispatch("track_exception", excpayload)

                elif data["type"] == "TrackStuckEvent":
                    track: Playable = Playable(data["track"])
                    threshold: int = data["thresholdMs"]

                    stuckpayload: TrackStuckEventPayload = TrackStuckEventPayload(
                        player=player, track=track, threshold=threshold
                    )
                    self.dispatch("track_stuck", stuckpayload)

                elif data["type"] == "WebSocketClosedEvent":
                    code: int = data["code"]
                    reason: str = data["reason"]
                    by_remote: bool = data["byRemote"]

                    wcpayload: WebsocketClosedEventPayload = WebsocketClosedEventPayload(
                        player=player, code=code, reason=reason, by_remote=by_remote
                    )
                    self.dispatch("websocket_closed", wcpayload)

                else:
                    other_payload: ExtraEventPayload = ExtraEventPayload(node=self.node, player=player, data=data)
                    self.dispatch("extra_event", other_payload)
            else:
                logger.debug("'Received an unknown OP from Lavalink '%s'. Disregarding.", data["op"])

    def get_player(self, guild_id: str | int) -> Player | None:
        return self.node.get_player(int(guild_id))

    def dispatch(self, event: str, /, *args: Any, **kwargs: Any) -> None:
        assert self.node.client is not None

        self.node.client.dispatch(f"wavelink_{event}", *args, **kwargs)
        logger.debug("%r dispatched the event 'on_wavelink_%s'", self.node, event)

    async def cleanup(self) -> None:
        if self.keep_alive_task:
            try:
                self.keep_alive_task.cancel()
            except Exception:
                pass

        if self.socket:
            try:
                await self.socket.close()
            except Exception:
                pass

        self.node._status = NodeStatus.DISCONNECTED
        self.node._session_id = None
        self.node._players = {}

        self.node._websocket = None

        logger.debug("Successfully cleaned up the websocket for %r", self.node)



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import logging
import secrets
import urllib.parse
from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypeAlias

import aiohttp
from discord.utils import classproperty

from . import __version__
from .enums import NodeStatus
from .exceptions import (
    AuthorizationFailedException,
    InvalidClientException,
    InvalidNodeException,
    LavalinkException,
    LavalinkLoadException,
    NodeException,
)
from .lfu import LFUCache
from .payloads import *
from .tracks import Playable, Playlist
from .websocket import Websocket


if TYPE_CHECKING:
    from collections.abc import Iterable

    import discord

    from .player import Player
    from .types.request import Request, UpdateSessionRequest
    from .types.response import (
        EmptyLoadedResponse,
        ErrorLoadedResponse,
        ErrorResponse,
        InfoResponse,
        PlayerResponse,
        PlaylistLoadedResponse,
        SearchLoadedResponse,
        StatsResponse,
        TrackLoadedResponse,
        UpdateResponse,
    )
    from .types.tracks import TrackPayload

    LoadedResponse: TypeAlias = (
        TrackLoadedResponse | SearchLoadedResponse | PlaylistLoadedResponse | EmptyLoadedResponse | ErrorLoadedResponse
    )


__all__ = ("Node", "Pool")


logger: logging.Logger = logging.getLogger(__name__)


Method = Literal["GET", "POST", "PATCH", "DELETE", "PUT", "OPTIONS"]


class Node:
    """The Node represents a connection to Lavalink.

    The Node is responsible for keeping the websocket alive, resuming session, sending API requests and keeping track
    of connected all :class:`~wavelink.Player`.

    .. container:: operations

        .. describe:: node == other

            Equality check to determine whether this Node is equal to another reference of a Node.

        .. describe:: repr(node)

            The official string representation of this Node.

    Parameters
    ----------
    identifier: str | None
        A unique identifier for this Node. Could be ``None`` to generate a random one on creation.
    uri: str
        The URL/URI that wavelink will use to connect to Lavalink. Usually this is in the form of something like:
        ``http://localhost:2333`` which includes the port. But you could also provide a domain which won't require a
        port like ``https://lavalink.example.com`` or a public IP address and port like ``http://111.333.444.55:2333``.
    password: str
        The password used to connect and authorize this Node.
    session: aiohttp.ClientSession | None
        An optional :class:`aiohttp.ClientSession` used to connect this Node over websocket and REST.
        If ``None``, one will be generated for you. Defaults to ``None``.
    heartbeat: Optional[float]
        A ``float`` in seconds to ping your websocket keep alive. Usually you would not change this.
    retries: int | None
        A ``int`` of retries to attempt when connecting or reconnecting this Node. When the retries are exhausted
        the Node will be closed and cleaned-up. ``None`` will retry forever. Defaults to ``None``.
    client: :class:`discord.Client` | None
        The :class:`discord.Client` or subclasses, E.g. ``commands.Bot`` used to connect this Node. If this is *not*
        passed you must pass this to :meth:`wavelink.Pool.connect`.
    resume_timeout: Optional[int]
        The seconds this Node should configure Lavalink for resuming its current session in case of network issues.
        If this is ``0`` or below, resuming will be disabled. Defaults to ``60``.
    inactive_player_timeout: int | None
        Set the default for :attr:`wavelink.Player.inactive_timeout` on every player that connects to this node.
        Defaults to ``300``.
    inactive_channel_tokens: int | None
        Sets the default for :attr:`wavelink.Player.inactive_channel_tokens` on every player that connects to this node.
        Defaults to ``3``.

        See also: :func:`on_wavelink_inactive_player`.
    """

    def __init__(
        self,
        *,
        identifier: str | None = None,
        uri: str,
        password: str,
        session: aiohttp.ClientSession | None = None,
        heartbeat: float = 15.0,
        retries: int | None = None,
        client: discord.Client | None = None,
        resume_timeout: int = 60,
        inactive_player_timeout: int | None = 300,
        inactive_channel_tokens: int | None = 3,
    ) -> None:
        self._identifier = identifier or secrets.token_urlsafe(12)
        self._uri = uri.removesuffix("/")
        self._password = password
        self._session = session or aiohttp.ClientSession()
        self._heartbeat = heartbeat
        self._retries = retries
        self._client = client
        self._resume_timeout = resume_timeout

        self._status: NodeStatus = NodeStatus.DISCONNECTED
        self._has_closed: bool = False
        self._session_id: str | None = None

        self._players: dict[int, Player] = {}
        self._total_player_count: int | None = None

        self._spotify_enabled: bool = False

        self._websocket: Websocket | None = None

        if inactive_player_timeout and inactive_player_timeout < 10:
            logger.warning('Setting "inactive_player_timeout" below 10 seconds may result in unwanted side effects.')

        self._inactive_player_timeout = (
            inactive_player_timeout if inactive_player_timeout and inactive_player_timeout > 0 else None
        )

        self._inactive_channel_tokens = inactive_channel_tokens

    def __repr__(self) -> str:
        return f"Node(identifier={self.identifier}, uri={self.uri}, status={self.status}, players={len(self.players)})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return NotImplemented

        return other.identifier == self.identifier

    @property
    def headers(self) -> dict[str, str]:
        """A property that returns the headers configured for sending API and websocket requests.

        .. warning::

            This includes your Node password. Please be vigilant when using this property.
        """
        assert self.client is not None
        assert self.client.user is not None

        data = {
            "Authorization": self.password,
            "User-Id": str(self.client.user.id),
            "Client-Name": f"Wavelink/{__version__}",
        }

        return data

    @property
    def identifier(self) -> str:
        """The unique identifier for this :class:`Node`.


        .. versionchanged:: 3.0.0

            This property was previously known as ``id``.
        """
        return self._identifier

    @property
    def uri(self) -> str:
        """The URI used to connect this :class:`Node` to Lavalink."""
        return self._uri

    @property
    def status(self) -> NodeStatus:
        """The current :class:`Node` status.

        Refer to: :class:`~wavelink.NodeStatus`
        """
        return self._status

    @property
    def players(self) -> dict[int, Player]:
        """A mapping of :attr:`discord.Guild.id` to :class:`~wavelink.Player`.


        .. versionchanged:: 3.1.1

            This property now returns a shallow copy of the internal mapping.
        """
        return self._players.copy()

    @property
    def client(self) -> discord.Client | None:
        """Returns the :class:`discord.Client` associated with this :class:`Node`.

        Could be ``None`` if it has not been set yet.


        .. versionadded:: 3.0.0
        """
        return self._client

    @property
    def password(self) -> str:
        """Returns the password used to connect this :class:`Node` to Lavalink.

        .. versionadded:: 3.0.0
        """
        return self._password

    @property
    def heartbeat(self) -> float:
        """Returns the duration in seconds that the :class:`Node` websocket should send a heartbeat.

        .. versionadded:: 3.0.0
        """
        return self._heartbeat

    @property
    def session_id(self) -> str | None:
        """Returns the Lavalink session ID. Could be None if this :class:`Node` has not connected yet.

        .. versionadded:: 3.0.0
        """
        return self._session_id

    async def _pool_closer(self) -> None:
        try:
            await self._session.close()
        except Exception:
            pass

        if not self._has_closed:
            await self.close()

    async def close(self, eject: bool = False) -> None:
        """Method to close this Node and cleanup.

        After this method has finished, the event ``on_wavelink_node_closed`` will be fired.

        This method renders the Node websocket disconnected and disconnects all players.

        Parameters
        ----------
        eject: bool
            If ``True``, this will remove the Node from the Pool. Defaults to ``False``.


        .. versionchanged:: 3.2.1

            Added the ``eject`` parameter. Fixed a bug where the connected Players were not being disconnected.
        """
        disconnected: list[Player] = []

        for player in self._players.copy().values():
            try:
                await player.disconnect()
            except Exception as e:
                logger.debug("An error occured while disconnecting a player in the close method of %r: %s", self, e)
                pass

            disconnected.append(player)

        if self._websocket is not None:
            await self._websocket.cleanup()

        self._status = NodeStatus.DISCONNECTED
        self._session_id = None
        self._players = {}

        self._has_closed = True

        if eject:
            getattr(Pool, "_Pool__nodes").pop(self.identifier, None)

        # Dispatch Node Closed Event... node, list of disconnected players
        if self.client is not None:
            self.client.dispatch("wavelink_node_closed", self, disconnected)

    async def _connect(self, *, client: discord.Client | None) -> None:
        client_ = self._client or client

        if not client_:
            raise InvalidClientException(f"Unable to connect {self!r} as you have not provided a valid discord.Client.")

        self._client = client_

        self._has_closed = False
        if not self._session or self._session.closed:
            self._session = aiohttp.ClientSession()

        websocket: Websocket = Websocket(node=self)
        self._websocket = websocket
        await websocket.connect()

    async def send(
        self, method: Method = "GET", *, path: str, data: Any | None = None, params: dict[str, Any] | None = None
    ) -> Any:
        """Method for making requests to the Lavalink node.

        .. warning::

            Usually you wouldn't use this method. Please use the built in methods of :class:`~Node`, :class:`~Pool`
            and :class:`~wavelink.Player`, unless you need to send specific plugin data to Lavalink.

            Using this method may have unwanted side effects on your players and/or nodes.

        Parameters
        ----------
        method: Optional[str]
            The method to use when making this request. Available methods are
            "GET", "POST", "PATCH", "PUT", "DELETE" and "OPTIONS". Defaults to "GET".
        path: str
            The path to make this request to. E.g. "/v4/stats".
        data: Any | None
            The optional JSON data to send along with your request to Lavalink. This should be a dict[str, Any]
            and able to be converted to JSON.
        params: Optional[dict[str, Any]]
            An optional dict of query parameters to send with your request to Lavalink. If you include your query
            parameters in the ``path`` parameter, do not pass them here as well. E.g. {"thing": 1, "other": 2}
            would equate to "?thing=1&other=2".

        Returns
        -------
        Any
            The response from Lavalink which will either be None, a str or JSON.

        Raises
        ------
        LavalinkException
            An error occurred while making this request to Lavalink.
        NodeException
            An error occured while making this request to Lavalink,
            and Lavalink was unable to send any error information.


        .. versionadded:: 3.0.0
        """
        clean_path: str = path.removesuffix("/")
        uri: str = f"{self.uri}/{clean_path}"

        if params is None:
            params = {}

        async with self._session.request(
            method=method, url=uri, params=params, json=data, headers=self.headers
        ) as resp:
            if resp.status == 204:
                return

            if resp.status >= 300:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

            try:
                rdata: Any = await resp.json()
            except aiohttp.ContentTypeError:
                pass
            else:
                return rdata

            try:
                body: str = await resp.text()
            except aiohttp.ClientError:
                return

            return body

    async def _fetch_players(self) -> list[PlayerResponse]:
        uri: str = f"{self.uri}/v4/sessions/{self.session_id}/players"

        async with self._session.get(url=uri, headers=self.headers) as resp:
            if resp.status == 200:
                resp_data: list[PlayerResponse] = await resp.json()
                return resp_data

            else:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

    async def fetch_players(self) -> list[PlayerResponsePayload]:
        """Method to fetch the player information Lavalink holds for every connected player on this node.

        .. warning::

            This payload is not the same as the :class:`wavelink.Player` class. This is the data received from
            Lavalink about the players.


        Returns
        -------
        list[:class:`PlayerResponsePayload`]
            A list of :class:`PlayerResponsePayload` representing each player connected to this node.

        Raises
        ------
        LavalinkException
            An error occurred while making this request to Lavalink.
        NodeException
            An error occured while making this request to Lavalink,
            and Lavalink was unable to send any error information.


        .. versionadded:: 3.1.0
        """
        data: list[PlayerResponse] = await self._fetch_players()

        payload: list[PlayerResponsePayload] = [PlayerResponsePayload(p) for p in data]
        return payload

    async def _fetch_player(self, guild_id: int, /) -> PlayerResponse:
        uri: str = f"{self.uri}/v4/sessions/{self.session_id}/players/{guild_id}"

        async with self._session.get(url=uri, headers=self.headers) as resp:
            if resp.status == 200:
                resp_data: PlayerResponse = await resp.json()
                return resp_data

            else:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

    async def fetch_player_info(self, guild_id: int, /) -> PlayerResponsePayload | None:
        """Method to fetch the player information Lavalink holds for the specific guild.

        .. warning::

            This payload is not the same as the :class:`wavelink.Player` class. This is the data received from
            Lavalink about the player. See: :meth:`~wavelink.Node.get_player`


        Parameters
        ----------
        guild_id: int
            The ID of the guild you want to fetch info for.

        Returns
        -------
        :class:`PlayerResponsePayload` | None
            The :class:`PlayerResponsePayload` representing the player info for the guild ID connected to this node.
            Could be ``None`` if no player is found with the given guild ID.

        Raises
        ------
        LavalinkException
            An error occurred while making this request to Lavalink.
        NodeException
            An error occured while making this request to Lavalink,
            and Lavalink was unable to send any error information.


        .. versionadded:: 3.1.0
        """
        try:
            data: PlayerResponse = await self._fetch_player(guild_id)
        except LavalinkException as e:
            if e.status == 404:
                return None

            raise e

        payload: PlayerResponsePayload = PlayerResponsePayload(data)
        return payload

    async def _update_player(self, guild_id: int, /, *, data: Request, replace: bool = False) -> PlayerResponse:
        no_replace: bool = not replace

        uri: str = f"{self.uri}/v4/sessions/{self.session_id}/players/{guild_id}?noReplace={no_replace}"

        async with self._session.patch(url=uri, json=data, headers=self.headers) as resp:
            if resp.status == 200:
                resp_data: PlayerResponse = await resp.json()
                return resp_data

            else:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

    async def _destroy_player(self, guild_id: int, /) -> None:
        uri: str = f"{self.uri}/v4/sessions/{self.session_id}/players/{guild_id}"

        async with self._session.delete(url=uri, headers=self.headers) as resp:
            if resp.status == 204:
                return

            try:
                exc_data: ErrorResponse = await resp.json()
            except Exception as e:
                logger.warning("An error occured making a request on %r: %s", self, e)
                raise NodeException(status=resp.status)

            raise LavalinkException(data=exc_data)

    async def _update_session(self, *, data: UpdateSessionRequest) -> UpdateResponse:
        uri: str = f"{self.uri}/v4/sessions/{self.session_id}"

        async with self._session.patch(url=uri, json=data, headers=self.headers) as resp:
            if resp.status == 200:
                resp_data: UpdateResponse = await resp.json()
                return resp_data

            else:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

    async def _fetch_tracks(self, query: str) -> LoadedResponse:
        uri: str = f"{self.uri}/v4/loadtracks?identifier={query}"

        async with self._session.get(url=uri, headers=self.headers) as resp:
            if resp.status == 200:
                resp_data: LoadedResponse = await resp.json()
                return resp_data

            else:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

    async def _decode_track(self) -> TrackPayload: ...

    async def _decode_tracks(self) -> list[TrackPayload]: ...

    async def _fetch_info(self) -> InfoResponse:
        uri: str = f"{self.uri}/v4/info"

        async with self._session.get(url=uri, headers=self.headers) as resp:
            if resp.status == 200:
                resp_data: InfoResponse = await resp.json()
                return resp_data

            else:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

    async def fetch_info(self) -> InfoResponsePayload:
        """Method to fetch this Lavalink Nodes info response data.

        Returns
        -------
        :class:`InfoResponsePayload`
            The :class:`InfoResponsePayload` associated with this Node.

        Raises
        ------
        LavalinkException
            An error occurred while making this request to Lavalink.
        NodeException
            An error occured while making this request to Lavalink,
            and Lavalink was unable to send any error information.


        .. versionadded:: 3.1.0
        """
        data: InfoResponse = await self._fetch_info()

        payload: InfoResponsePayload = InfoResponsePayload(data)
        return payload

    async def _fetch_stats(self) -> StatsResponse:
        uri: str = f"{self.uri}/v4/stats"

        async with self._session.get(url=uri, headers=self.headers) as resp:
            if resp.status == 200:
                resp_data: StatsResponse = await resp.json()
                return resp_data

            else:
                try:
                    exc_data: ErrorResponse = await resp.json()
                except Exception as e:
                    logger.warning("An error occured making a request on %r: %s", self, e)
                    raise NodeException(status=resp.status)

                raise LavalinkException(data=exc_data)

    async def fetch_stats(self) -> StatsResponsePayload:
        """Method to fetch this Lavalink Nodes stats response data.

        Returns
        -------
        :class:`StatsResponsePayload`
            The :class:`StatsResponsePayload` associated with this Node.

        Raises
        ------
        LavalinkException
            An error occurred while making this request to Lavalink.
        NodeException
            An error occured while making this request to Lavalink,
            and Lavalink was unable to send any error information.


        .. versionadded:: 3.1.0
        """
        data: StatsResponse = await self._fetch_stats()

        payload: StatsResponsePayload = StatsResponsePayload(data)
        return payload

    async def _fetch_version(self) -> str:
        uri: str = f"{self.uri}/version"

        async with self._session.get(url=uri, headers=self.headers) as resp:
            if resp.status == 200:
                return await resp.text()

            try:
                exc_data: ErrorResponse = await resp.json()
            except Exception as e:
                logger.warning("An error occured making a request on %r: %s", self, e)
                raise NodeException(status=resp.status)

            raise LavalinkException(data=exc_data)

    async def fetch_version(self) -> str:
        """Method to fetch this Lavalink version string.

        Returns
        -------
        str
            The version string associated with this Lavalink node.

        Raises
        ------
        LavalinkException
            An error occurred while making this request to Lavalink.
        NodeException
            An error occured while making this request to Lavalink,
            and Lavalink was unable to send any error information.


        .. versionadded:: 3.1.0
        """
        data: str = await self._fetch_version()
        return data

    def get_player(self, guild_id: int, /) -> Player | None:
        """Return a :class:`~wavelink.Player` associated with the provided :attr:`discord.Guild.id`.

        Parameters
        ----------
        guild_id: int
            The :attr:`discord.Guild.id` to retrieve a :class:`~wavelink.Player` for.

        Returns
        -------
        Optional[:class:`~wavelink.Player`]
            The Player associated with this guild ID. Could be None if no :class:`~wavelink.Player` exists
            for this guild.
        """
        return self._players.get(guild_id, None)


class Pool:
    """The wavelink Pool represents a collection of :class:`~wavelink.Node` and helper methods for searching tracks.

    To connect a :class:`~wavelink.Node` please use this Pool.

    .. note::

        All methods and attributes on this class are class level, not instance. Do not create an instance of this class.
    """

    __nodes: ClassVar[dict[str, Node]] = {}
    __cache: LFUCache | None = None

    @classmethod
    async def connect(
        cls, *, nodes: Iterable[Node], client: discord.Client | None = None, cache_capacity: int | None = None
    ) -> dict[str, Node]:
        """Connect the provided Iterable[:class:`Node`] to Lavalink.

        Parameters
        ----------
        nodes: Iterable[:class:`Node`]
            The :class:`Node`'s to connect to Lavalink.
        client: :class:`discord.Client` | None
            The :class:`discord.Client` to use to connect the :class:`Node`. If the Node already has a client
            set, this method will **not** override it. Defaults to None.
        cache_capacity: int | None
            An optional integer of the amount of track searches to cache. This is an experimental mode.
            Passing ``None`` will disable this experiment. Defaults to ``None``.

        Returns
        -------
        dict[str, :class:`Node`]
            A mapping of :attr:`Node.identifier` to :class:`Node` associated with the :class:`Pool`.


        Raises
        ------
        AuthorizationFailedException
            The node password was incorrect.
        InvalidClientException
            The :class:`discord.Client` passed was not valid.
        NodeException
            The node failed to connect properly. Please check that your Lavalink version is version 4.


        .. versionchanged:: 3.0.0

            The ``client`` parameter is no longer required.
            Added the ``cache_capacity`` parameter.
        """
        for node in nodes:
            client_ = node.client or client

            if node.identifier in cls.__nodes:
                msg: str = f'Unable to connect {node!r} as you already have a node with identifier "{node.identifier}"'
                logger.error(msg)

                continue

            if node.status in (NodeStatus.CONNECTING, NodeStatus.CONNECTED):
                logger.error("Unable to connect %r as it is already in a connecting or connected state.", node)
                continue

            try:
                await node._connect(client=client_)
            except InvalidClientException as e:
                logger.error(e)
            except AuthorizationFailedException:
                logger.error("Failed to authenticate %r on Lavalink with the provided password.", node)
            except NodeException:
                logger.error(
                    "Failed to connect to %r. Check that your Lavalink major version is '4' and that you are trying to connect to Lavalink on the correct port.",
                    node,
                )
            else:
                cls.__nodes[node.identifier] = node

        if cache_capacity is not None and cls.nodes:
            if cache_capacity <= 0:
                logger.warning("LFU Request cache capacity must be > 0. Not enabling cache.")

            else:
                cls.__cache = LFUCache(capacity=cache_capacity)
                logger.info("Experimental request caching has been toggled ON. To disable run Pool.toggle_cache()")

        return cls.nodes

    @classmethod
    async def reconnect(cls) -> dict[str, Node]:
        for node in cls.__nodes.values():
            if node.status is not NodeStatus.DISCONNECTED:
                continue

            try:
                await node._connect(client=None)
            except InvalidClientException as e:
                logger.error(e)
            except AuthorizationFailedException:
                logger.error("Failed to authenticate %r on Lavalink with the provided password.", node)
            except NodeException:
                logger.error(
                    "Failed to connect to %r. Check that your Lavalink major version is '4' and that you are trying to connect to Lavalink on the correct port.",
                    node,
                )

        return cls.nodes

    @classmethod
    async def close(cls) -> None:
        """Close and clean up all :class:`~wavelink.Node` on this Pool.

        This calls :meth:`wavelink.Node.close` on each node.


        .. versionadded:: 3.0.0
        """
        for node in cls.__nodes.values():
            await node.close()

    @classproperty
    def nodes(cls) -> dict[str, Node]:
        """A mapping of :attr:`Node.identifier` to :class:`Node` that have previously been successfully connected.


        .. versionchanged:: 3.0.0

            This property now returns a copy.
        """
        nodes = cls.__nodes.copy()
        return nodes

    @classmethod
    def get_node(cls, identifier: str | None = None, /) -> Node:
        """Retrieve a :class:`Node` from the :class:`Pool` with the given identifier.

        If no identifier is provided, this method returns the ``best`` node.

        Parameters
        ----------
        identifier: str | None
            An optional identifier to retrieve a :class:`Node`.

        Raises
        ------
        InvalidNodeException
            Raised when a Node can not be found, or no :class:`Node` exists on the :class:`Pool`.


        .. versionchanged:: 3.0.0

            The ``id`` parameter was changed to ``identifier`` and is positional only.
        """
        if identifier:
            if identifier not in cls.__nodes:
                raise InvalidNodeException(f'A Node with the identifier "{identifier}" does not exist.')

            return cls.__nodes[identifier]

        nodes: list[Node] = [n for n in cls.__nodes.values() if n.status is NodeStatus.CONNECTED]
        if not nodes:
            raise InvalidNodeException("No nodes are currently assigned to the wavelink.Pool in a CONNECTED state.")

        return sorted(nodes, key=lambda n: n._total_player_count or len(n.players))[0]

    @classmethod
    async def fetch_tracks(cls, query: str, /, *, node: Node | None = None) -> list[Playable] | Playlist:
        """Search for a list of :class:`~wavelink.Playable` or a :class:`~wavelink.Playlist`, with the given query.

        Parameters
        ----------
        query: str
            The query to search tracks for. If this is not a URL based search you should provide the appropriate search
            prefix, e.g. "ytsearch:Rick Roll"
        node: :class:`~wavelink.Node` | None
            An optional :class:`~wavelink.Node` to use when fetching tracks. Defaults to ``None``, which selects the
            most appropriate :class:`~wavelink.Node` automatically.

        Returns
        -------
        list[Playable] | Playlist
            A list of :class:`~wavelink.Playable` or a :class:`~wavelink.Playlist`
            based on your search ``query``. Could be an empty list, if no tracks were found.

        Raises
        ------
        LavalinkLoadException
            Exception raised when Lavalink fails to load results based on your query.


        .. versionchanged:: 3.0.0

            This method was previously known as both ``.get_tracks`` and ``.get_playlist``. This method now searches
            for both :class:`~wavelink.Playable` and :class:`~wavelink.Playlist` and returns the appropriate type,
            or an empty list if no results were found.

            This method no longer accepts the ``cls`` parameter.


        .. versionadded:: 3.4.0

            Added the ``node`` Keyword-Only argument.
        """

        # TODO: Documentation Extension for `.. positional-only::` marker.
        encoded_query: str = urllib.parse.quote(query)

        if cls.__cache is not None:
            potential: list[Playable] | Playlist = cls.__cache.get(encoded_query, None)

            if potential:
                return potential

        node_: Node = node or cls.get_node()
        resp: LoadedResponse = await node_._fetch_tracks(encoded_query)

        if resp["loadType"] == "track":
            track = Playable(data=resp["data"])

            if cls.__cache is not None and not track.is_stream:
                cls.__cache.put(encoded_query, [track])

            return [track]

        elif resp["loadType"] == "search":
            tracks = [Playable(data=tdata) for tdata in resp["data"]]

            if cls.__cache is not None:
                cls.__cache.put(encoded_query, tracks)

            return tracks

        if resp["loadType"] == "playlist":
            playlist: Playlist = Playlist(data=resp["data"])

            if cls.__cache is not None:
                cls.__cache.put(encoded_query, playlist)

            return playlist

        elif resp["loadType"] == "empty":
            return []

        elif resp["loadType"] == "error":
            raise LavalinkLoadException(data=resp["data"])

        else:
            return []

    @classmethod
    def cache(cls, capacity: int | None | bool = None) -> None:
        if capacity in (None, False) or capacity <= 0:
            cls.__cache = None
            return

        if not isinstance(capacity, int):  # type: ignore
            raise ValueError("The LFU cache expects an integer, None or bool.")

        cls.__cache = LFUCache(capacity=capacity)

    @classmethod
    def has_cache(cls) -> bool:
        return cls.__cache is not None



import argparse
import platform
import subprocess
import sys

import wavelink


parser = argparse.ArgumentParser(prog="wavelink")
parser.add_argument("--version", action="store_true", help="Get version and debug information for wavelink.")


args = parser.parse_args()


def get_debug_info() -> None:
    python_info = "\n".join(sys.version.split("\n"))
    java_version = subprocess.check_output(["java", "-version"], stderr=subprocess.STDOUT)
    java_version = f'\n{" " * 8}- '.join(v for v in java_version.decode().split("\r\n") if v)

    info: str = f"""
    wavelink: {wavelink.__version__}

    Python:
        - {python_info}
    System:
        - {platform.platform()}
    Java:
        - {java_version or "Version Not Found"}
    """

    print(info)


if args.version:
    get_debug_info()



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Literal, TypedDict

from typing_extensions import Never, NotRequired

from .filters import FilterPayload
from .state import PlayerState
from .stats import CPUStats, FrameStats, MemoryStats
from .tracks import PlaylistPayload, TrackPayload


class ErrorResponse(TypedDict):
    timestamp: int
    status: int
    error: str
    trace: NotRequired[str]
    message: str
    path: str


class LoadedErrorPayload(TypedDict):
    message: str
    severity: str
    cause: str


class VoiceStateResponse(TypedDict, total=False):
    token: str
    endpoint: str | None
    sessionId: str


class PlayerResponse(TypedDict):
    guildId: str
    track: NotRequired[TrackPayload]
    volume: int
    paused: bool
    state: PlayerState
    voice: VoiceStateResponse
    filters: FilterPayload


class UpdateResponse(TypedDict):
    resuming: bool
    timeout: int


class TrackLoadedResponse(TypedDict):
    loadType: Literal["track"]
    data: TrackPayload


class PlaylistLoadedResponse(TypedDict):
    loadType: Literal["playlist"]
    data: PlaylistPayload


class SearchLoadedResponse(TypedDict):
    loadType: Literal["search"]
    data: list[TrackPayload]


class EmptyLoadedResponse(TypedDict):
    loadType: Literal["empty"]
    data: dict[Never, Never]


class ErrorLoadedResponse(TypedDict):
    loadType: Literal["error"]
    data: LoadedErrorPayload


class VersionPayload(TypedDict):
    semver: str
    major: int
    minor: int
    patch: int
    preRelease: NotRequired[str]
    build: NotRequired[str]


class GitPayload(TypedDict):
    branch: str
    commit: str
    commitTime: int


class PluginPayload(TypedDict):
    name: str
    version: str


class InfoResponse(TypedDict):
    version: VersionPayload
    buildTime: int
    git: GitPayload
    jvm: str
    lavaplayer: str
    sourceManagers: list[str]
    filters: list[str]
    plugins: list[PluginPayload]


class StatsResponse(TypedDict):
    players: int
    playingPlayers: int
    uptime: int
    memory: MemoryStats
    cpu: CPUStats
    frameStats: NotRequired[FrameStats]



"""
The MIT License (MIT)

Copyright (c) 2021-Present PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypedDict


if TYPE_CHECKING:
    from typing_extensions import Self, Unpack

    from .types.filters import (
        ChannelMix as ChannelMixPayload,
        Distortion as DistortionPayload,
        Equalizer as EqualizerPayload,
        FilterPayload,
        Karaoke as KaraokePayload,
        LowPass as LowPassPayload,
        Rotation as RotationPayload,
        Timescale as TimescalePayload,
        Tremolo as TremoloPayload,
        Vibrato as VibratoPayload,
    )


__all__ = (
    "FiltersOptions",
    "Filters",
    "Equalizer",
    "Karaoke",
    "Timescale",
    "Tremolo",
    "Vibrato",
    "Rotation",
    "Distortion",
    "ChannelMix",
    "LowPass",
    "PluginFilters",
)


class FiltersOptions(TypedDict, total=False):
    volume: float
    equalizer: Equalizer
    karaoke: Karaoke
    timescale: Timescale
    tremolo: Tremolo
    vibrato: Vibrato
    rotation: Rotation
    distortion: Distortion
    channel_mix: ChannelMix
    low_pass: LowPass
    plugin_filters: PluginFilters
    reset: bool


class EqualizerOptions(TypedDict):
    bands: list[EqualizerPayload] | None


class KaraokeOptions(TypedDict):
    level: float | None
    mono_level: float | None
    filter_band: float | None
    filter_width: float | None


class RotationOptions(TypedDict):
    rotation_hz: float | None


class DistortionOptions(TypedDict):
    sin_offset: float | None
    sin_scale: float | None
    cos_offset: float | None
    cos_scale: float | None
    tan_offset: float | None
    tan_scale: float | None
    offset: float | None
    scale: float | None


class ChannelMixOptions(TypedDict):
    left_to_left: float | None
    left_to_right: float | None
    right_to_left: float | None
    right_to_right: float | None


class Equalizer:
    """Equalizer Filter Class.

    There are 15 bands ``0`` to ``14`` that can be changed.
    Each band has a ``gain`` which is the multiplier for the given band. ``gain`` defaults to ``0``.

    Valid ``gain`` values range from ``-0.25`` to ``1.0``, where ``-0.25`` means the given band is completely muted,
    and ``0.25`` means it will be doubled.

    Modifying the ``gain`` could also change the volume of the output.
    """

    def __init__(self, payload: list[EqualizerPayload] | None = None) -> None:
        if payload and len(payload) == 15:
            self._payload = self._set(payload)

        else:
            payload_: dict[int, EqualizerPayload] = {n: {"band": n, "gain": 0.0} for n in range(15)}
            self._payload = payload_

    def _set(self, payload: list[EqualizerPayload]) -> dict[int, EqualizerPayload]:
        default: dict[int, EqualizerPayload] = {n: {"band": n, "gain": 0.0} for n in range(15)}

        for eq in payload:
            band: int = eq["band"]
            if band > 14 or band < 0:
                continue

            default[band] = eq

        return default

    def set(self, **options: Unpack[EqualizerOptions]) -> Self:
        """Set the bands of the Equalizer class.

        Accepts a keyword argument ``bands`` which is a ``list`` of ``dict`` containing the keys ``band`` and ``gain``.

        ``band`` can be an ``int`` beteween ``0`` and ``14``.
        ``gain`` can be a float between ``-0.25`` and ``1.0``, where ``-0.25`` means the given band is completely muted,
        and ``0.25`` means it will be doubled.

        Using this method changes **all** bands, resetting any bands not provided.
        To change specific bands, consider accessing :attr:`~wavelink.Equalizer.payload` first.
        """
        default: dict[int, EqualizerPayload] = {n: {"band": n, "gain": 0.0} for n in range(15)}
        payload: list[EqualizerPayload] | None = options.get("bands", None)

        if payload is None:
            self._payload = default
            return self

        self._payload = self._set(payload)
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: dict[int, EqualizerPayload] = {n: {"band": n, "gain": 0.0} for n in range(15)}
        return self

    @property
    def payload(self) -> dict[int, EqualizerPayload]:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "Equalizer"

    def __repr__(self) -> str:
        return f"<Equalizer: {self._payload}>"


class Karaoke:
    """Karaoke Filter class.

    Uses equalization to eliminate part of a band, usually targeting vocals.
    """

    def __init__(self, payload: KaraokePayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[KaraokeOptions]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        level: Optional[float]
            The level ``0`` to ``1.0`` where ``0.0`` is no effect and ``1.0`` is full effect.
        mono_level: Optional[float]
            The mono level ``0`` to ``1.0`` where ``0.0`` is no effect and ``1.0`` is full effect.
        filter_band: Optional[float]
            The filter band in Hz.
        filter_width: Optional[float]
            The filter width.
        """
        self._payload: KaraokePayload = {
            "level": options.get("level", self._payload.get("level")),
            "monoLevel": options.get("mono_level", self._payload.get("monoLevel")),
            "filterBand": options.get("filter_band", self._payload.get("filterBand")),
            "filterWidth": options.get("filter_width", self._payload.get("filterWidth")),
        }
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: KaraokePayload = {}
        return self

    @property
    def payload(self) -> KaraokePayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "Karaoke"

    def __repr__(self) -> str:
        return f"<Karaoke: {self._payload}>"


class Timescale:
    """Timescale Filter class.

    Changes the speed, pitch, and rate.
    """

    def __init__(self, payload: TimescalePayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[TimescalePayload]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        speed: Optional[float]
            The playback speed.
        pitch: Optional[float]
            The pitch.
        rate: Optional[float]
            The rate.
        """
        self._payload.update(options)
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: TimescalePayload = {}
        return self

    @property
    def payload(self) -> TimescalePayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "Timescale"

    def __repr__(self) -> str:
        return f"<Timescale: {self._payload}>"


class Tremolo:
    """The Tremolo Filter class.

    Uses amplification to create a shuddering effect, where the volume quickly oscillates.
    Demo: https://en.wikipedia.org/wiki/File:Fuse_Electronics_Tremolo_MK-III_Quick_Demo.ogv
    """

    def __init__(self, payload: TremoloPayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[TremoloPayload]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        frequency: Optional[float]
            The frequency.
        depth: Optional[float]
            The tremolo depth.
        """
        self._payload.update(options)
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: TremoloPayload = {}
        return self

    @property
    def payload(self) -> TremoloPayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "Tremolo"

    def __repr__(self) -> str:
        return f"<Tremolo: {self._payload}>"


class Vibrato:
    """The Vibrato Filter class.

    Similar to tremolo. While tremolo oscillates the volume, vibrato oscillates the pitch.
    """

    def __init__(self, payload: VibratoPayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[VibratoPayload]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        frequency: Optional[float]
            The frequency.
        depth: Optional[float]
            The vibrato depth.
        """
        self._payload.update(options)
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: VibratoPayload = {}
        return self

    @property
    def payload(self) -> VibratoPayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "Vibrato"

    def __repr__(self) -> str:
        return f"<Vibrato: {self._payload}>"


class Rotation:
    """The Rotation Filter class.

    Rotates the sound around the stereo channels/user headphones (aka Audio Panning).
    It can produce an effect similar to https://youtu.be/QB9EB8mTKcc (without the reverb).
    """

    def __init__(self, payload: RotationPayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[RotationOptions]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        rotation_hz: Optional[float]
            The frequency of the audio rotating around the listener in Hz. ``0.2`` is similar to the example video.
        """
        self._payload: RotationPayload = {"rotationHz": options.get("rotation_hz", self._payload.get("rotationHz"))}
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: RotationPayload = {}
        return self

    @property
    def payload(self) -> RotationPayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "Rotation"

    def __repr__(self) -> str:
        return f"<Rotation: {self._payload}>"


class Distortion:
    """The Distortion Filter class.

    According to Lavalink "It can generate some pretty unique audio effects."
    """

    def __init__(self, payload: DistortionPayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[DistortionOptions]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        sin_offset: Optional[float]
            The sin offset.
        sin_scale: Optional[float]
            The sin scale.
        cos_offset: Optional[float]
            The cos offset.
        cos_scale: Optional[float]
            The cos scale.
        tan_offset: Optional[float]
            The tan offset.
        tan_scale: Optional[float]
            The tan scale.
        offset: Optional[float]
            The offset.
        scale: Optional[float]
            The scale.
        """
        self._payload: DistortionPayload = {
            "sinOffset": options.get("sin_offset", self._payload.get("sinOffset")),
            "sinScale": options.get("sin_scale", self._payload.get("sinScale")),
            "cosOffset": options.get("cos_offset", self._payload.get("cosOffset")),
            "cosScale": options.get("cos_scale", self._payload.get("cosScale")),
            "tanOffset": options.get("tan_offset", self._payload.get("tanOffset")),
            "tanScale": options.get("tan_scale", self._payload.get("tanScale")),
            "offset": options.get("offset", self._payload.get("offset")),
            "scale": options.get("scale", self._payload.get("scale")),
        }
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: DistortionPayload = {}
        return self

    @property
    def payload(self) -> DistortionPayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "Distortion"

    def __repr__(self) -> str:
        return f"<Distortion: {self._payload}>"


class ChannelMix:
    """The ChannelMix Filter class.

    Mixes both channels (left and right), with a configurable factor on how much each channel affects the other.
    With the defaults, both channels are kept independent of each other.

    Setting all factors to ``0.5`` means both channels get the same audio.
    """

    def __init__(self, payload: ChannelMixPayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[ChannelMixOptions]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        left_to_left: Optional[float]
            The left to left channel mix factor. Between ``0.0`` and ``1.0``.
        left_to_right: Optional[float]
            The left to right channel mix factor. Between ``0.0`` and ``1.0``.
        right_to_left: Optional[float]
            The right to left channel mix factor. Between ``0.0`` and ``1.0``.
        right_to_right: Optional[float]
            The right to right channel mix factor. Between ``0.0`` and ``1.0``.
        """
        self._payload: ChannelMixPayload = {
            "leftToLeft": options.get("left_to_left", self._payload.get("leftToLeft")),
            "leftToRight": options.get("left_to_right", self._payload.get("leftToRight")),
            "rightToLeft": options.get("right_to_left", self._payload.get("rightToLeft")),
            "rightToRight": options.get("right_to_right", self._payload.get("rightToRight")),
        }
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: ChannelMixPayload = {}
        return self

    @property
    def payload(self) -> ChannelMixPayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "ChannelMix"

    def __repr__(self) -> str:
        return f"<ChannelMix: {self._payload}>"


class LowPass:
    """The LowPass Filter class.

    Higher frequencies get suppressed, while lower frequencies pass through this filter, thus the name low pass.
    Any smoothing values equal to or less than ``1.0`` will disable the filter.
    """

    def __init__(self, payload: LowPassPayload) -> None:
        self._payload = payload

    def set(self, **options: Unpack[LowPassPayload]) -> Self:
        """Set the properties of the this filter.

        This method accepts keyword argument pairs.
        This method does not override existing settings if they are not provided.

        Parameters
        ----------
        smoothing: Optional[float]
            The smoothing factor.
        """
        self._payload.update(options)
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: LowPassPayload = {}
        return self

    @property
    def payload(self) -> LowPassPayload:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "LowPass"

    def __repr__(self) -> str:
        return f"<LowPass: {self._payload}>"


class PluginFilters:
    """The PluginFilters class.

    This class handles setting filters on plugins that support setting filter values.
    See the documentation of the Lavalink Plugin for more information on the values that can be set.

    This class takes in a ``dict[str, Any]`` usually in the form of:

    .. code:: python3

        {"pluginName": {"filterKey": "filterValue"}, ...}


    .. warning::

        Do NOT include the ``"pluginFilters"`` top level key when setting your values for this class.
    """

    def __init__(self, payload: dict[str, Any]) -> None:
        self._payload = payload

    def set(self, **options: dict[str, Any]) -> Self:
        """Set the properties of this filter.

        This method accepts keyword argument pairs OR you can alternatively unpack a dictionary.
        See the documentation of the Lavalink Plugin for more information on the values that can be set.

        Examples
        --------

        .. code:: python3

            plugin_filters: PluginFilters = PluginFilters()
            plugin_filters.set(pluginName={"filterKey": "filterValue", ...})

            # OR...

            plugin_filters.set(**{"pluginName": {"filterKey": "filterValue", ...}})
        """
        self._payload.update(options)
        return self

    def reset(self) -> Self:
        """Reset this filter to its defaults."""
        self._payload: dict[str, Any] = {}
        return self

    @property
    def payload(self) -> dict[str, Any]:
        """The raw payload associated with this filter.

        This property returns a copy.
        """
        return self._payload.copy()

    def __str__(self) -> str:
        return "PluginFilters"

    def __repr__(self) -> str:
        return f"<PluginFilters: {self._payload}"


class Filters:
    """The wavelink Filters class.

    This class contains the information associated with each of Lavalinks filter objects, as Python classes.
    Each filter can be ``set`` or ``reset`` individually.

    Using ``set`` on an individual filter only updates any ``new`` values you pass.
    Using ``reset`` on an individual filter, resets it's payload, and can be used before ``set`` when you want a clean
    state for that filter.

    See: :meth:`~wavelink.Filters.reset` to reset **every** individual filter.

    This class is already applied an instantiated on all new :class:`~wavelink.Player`.

    See: :meth:`~wavelink.Player.set_filters` for information on applying this class to your :class:`~wavelink.Player`.
    See: :attr:`~wavelink.Player.filters` for retrieving the applied filters.

    To retrieve the ``payload`` for this Filters class, you can call an instance of this class.

    Examples
    --------

    .. code:: python3

        import wavelink

        # Create a brand new Filters and apply it...
        # You can use player.set_filters() for an easier way to reset.
        filters: wavelink.Filters = wavelink.Filters()
        await player.set_filters(filters)


        # Retrieve the payload of any Filters instance...
        filters: wavelink.Filters = player.filters
        print(filters())


        # Set some filters...
        # You can set and reset individual filters at the same time...
        filters: wavelink.Filters = player.filters
        filters.timescale.set(pitch=1.2, speed=1.1, rate=1)
        filters.rotation.set(rotation_hz=0.2)
        filters.equalizer.reset()

        await player.set_filters(filters)


        # Reset a filter...
        filters: wavelink.Filters = player.filters
        filters.timescale.reset()

        await player.set_filters(filters)


        # Reset all filters...
        filters: wavelink.Filters = player.filters
        filters.reset()

        await player.set_filters(filters)


        # Reset and apply filters easier method...
        await player.set_filters()
    """

    def __init__(self, *, data: FilterPayload | None = None) -> None:
        self._volume: float | None = None
        self._equalizer: Equalizer = Equalizer(None)
        self._karaoke: Karaoke = Karaoke({})
        self._timescale: Timescale = Timescale({})
        self._tremolo: Tremolo = Tremolo({})
        self._vibrato: Vibrato = Vibrato({})
        self._rotation: Rotation = Rotation({})
        self._distortion: Distortion = Distortion({})
        self._channel_mix: ChannelMix = ChannelMix({})
        self._low_pass: LowPass = LowPass({})
        self._plugin_filters: PluginFilters = PluginFilters({})

        if data:
            self._create_from(data)

    def _create_from(self, data: FilterPayload) -> None:
        self._volume = data.get("volume")
        self._equalizer = Equalizer(data.get("equalizer", None))
        self._karaoke = Karaoke(data.get("karaoke", {}))
        self._timescale = Timescale(data.get("timescale", {}))
        self._tremolo = Tremolo(data.get("tremolo", {}))
        self._vibrato = Vibrato(data.get("vibrato", {}))
        self._rotation = Rotation(data.get("rotation", {}))
        self._distortion = Distortion(data.get("distortion", {}))
        self._channel_mix = ChannelMix(data.get("channelMix", {}))
        self._low_pass = LowPass(data.get("lowPass", {}))
        self._plugin_filters = PluginFilters(data.get("pluginFilters", {}))

    def _set_with_reset(self, filters: FiltersOptions) -> None:
        self._volume = filters.get("volume")
        self._equalizer = filters.get("equalizer", Equalizer(None))
        self._karaoke = filters.get("karaoke", Karaoke({}))
        self._timescale = filters.get("timescale", Timescale({}))
        self._tremolo = filters.get("tremolo", Tremolo({}))
        self._vibrato = filters.get("vibrato", Vibrato({}))
        self._rotation = filters.get("rotation", Rotation({}))
        self._distortion = filters.get("distortion", Distortion({}))
        self._channel_mix = filters.get("channel_mix", ChannelMix({}))
        self._low_pass = filters.get("low_pass", LowPass({}))
        self._plugin_filters = filters.get("plugin_filters", PluginFilters({}))

    def set_filters(self, **filters: Unpack[FiltersOptions]) -> None:
        """Set multiple filters at once to a standalone Filter object.
        To set the filters to the player directly see :meth:`wavelink.Player.set_filters`

        Parameters
        ----------
        volume: float
            The Volume filter to apply to the player.
        equalizer: :class:`wavelink.Equalizer`
            The Equalizer filter to apply to the player.
        karaoke: :class:`wavelink.Karaoke`
            The Karaoke filter to apply to the player.
        timescale: :class:`wavelink.Timescale`
            The Timescale filter to apply to the player.
        tremolo: :class:`wavelink.Tremolo`
            The Tremolo filter to apply to the player.
        vibrato: :class:`wavelink.Vibrato`
            The Vibrato filter to apply to the player.
        rotation: :class:`wavelink.Rotation`
            The Rotation filter to apply to the player.
        distortion: :class:`wavelink.Distortion`
            The Distortion filter to apply to the player.
        channel_mix: :class:`wavelink.ChannelMix`
            The ChannelMix filter to apply to the player.
        low_pass: :class:`wavelink.LowPass`
            The LowPass filter to apply to the player.
        plugin_filters: :class:`wavelink.PluginFilters`
            The extra Plugin Filters to apply to the player. See :class:`~wavelink.PluginFilters` for more details.
        reset: bool
            Whether to reset all filters that were not specified.
        """

        reset: bool = filters.get("reset", False)
        if reset:
            self._set_with_reset(filters)
            return

        self._volume = filters.get("volume", self._volume)
        self._equalizer = filters.get("equalizer", self._equalizer)
        self._karaoke = filters.get("karaoke", self._karaoke)
        self._timescale = filters.get("timescale", self._timescale)
        self._tremolo = filters.get("tremolo", self._tremolo)
        self._vibrato = filters.get("vibrato", self._vibrato)
        self._rotation = filters.get("rotation", self._rotation)
        self._distortion = filters.get("distortion", self._distortion)
        self._channel_mix = filters.get("channel_mix", self._channel_mix)
        self._low_pass = filters.get("low_pass", self._low_pass)
        self._plugin_filters = filters.get("plugin_filters", self._plugin_filters)

    def _reset(self) -> None:
        self._volume = None
        self._equalizer = Equalizer(None)
        self._karaoke = Karaoke({})
        self._timescale = Timescale({})
        self._tremolo = Tremolo({})
        self._vibrato = Vibrato({})
        self._rotation = Rotation({})
        self._distortion = Distortion({})
        self._channel_mix = ChannelMix({})
        self._low_pass = LowPass({})
        self._plugin_filters = PluginFilters({})

    def reset(self) -> None:
        """Method which resets this object to an original state.

        This method will clear all individual filters, and assign the wavelink default classes.
        """
        self._reset()

    @classmethod
    def from_filters(cls, **filters: Unpack[FiltersOptions]) -> Self:
        """Creates a Filters object with specified filters.

        Parameters
        ----------
        volume: float
            The Volume filter to apply to the player.
        equalizer: :class:`wavelink.Equalizer`
            The Equalizer filter to apply to the player.
        karaoke: :class:`wavelink.Karaoke`
            The Karaoke filter to apply to the player.
        timescale: :class:`wavelink.Timescale`
            The Timescale filter to apply to the player.
        tremolo: :class:`wavelink.Tremolo`
            The Tremolo filter to apply to the player.
        vibrato: :class:`wavelink.Vibrato`
            The Vibrato filter to apply to the player.
        rotation: :class:`wavelink.Rotation`
            The Rotation filter to apply to the player.
        distortion: :class:`wavelink.Distortion`
            The Distortion filter to apply to the player.
        channel_mix: :class:`wavelink.ChannelMix`
            The ChannelMix filter to apply to the player.
        low_pass: :class:`wavelink.LowPass`
            The LowPass filter to apply to the player.
        plugin_filters: :class:`wavelink.PluginFilters`
            The extra Plugin Filters to apply to the player. See :class:`~wavelink.PluginFilters` for more details.
        reset: bool
            Whether to reset all filters that were not specified.
        """

        self = cls()
        self._set_with_reset(filters)

        return self

    @property
    def volume(self) -> float | None:
        """Property which returns the volume ``float`` associated with this Filters payload.

        Adjusts the player volume from 0.0 to 5.0, where 1.0 is 100%. Values >1.0 may cause clipping.
        """
        return self._volume

    @volume.setter
    def volume(self, value: float) -> None:
        self._volume = value

    @property
    def equalizer(self) -> Equalizer:
        """Property which returns the :class:`~wavelink.Equalizer` filter associated with this Filters payload."""
        return self._equalizer

    @property
    def karaoke(self) -> Karaoke:
        """Property which returns the :class:`~wavelink.Karaoke` filter associated with this Filters payload."""
        return self._karaoke

    @property
    def timescale(self) -> Timescale:
        """Property which returns the :class:`~wavelink.Timescale` filter associated with this Filters payload."""
        return self._timescale

    @property
    def tremolo(self) -> Tremolo:
        """Property which returns the :class:`~wavelink.Tremolo` filter associated with this Filters payload."""
        return self._tremolo

    @property
    def vibrato(self) -> Vibrato:
        """Property which returns the :class:`~wavelink.Vibrato` filter associated with this Filters payload."""
        return self._vibrato

    @property
    def rotation(self) -> Rotation:
        """Property which returns the :class:`~wavelink.Rotation` filter associated with this Filters payload."""
        return self._rotation

    @property
    def distortion(self) -> Distortion:
        """Property which returns the :class:`~wavelink.Distortion` filter associated with this Filters payload."""
        return self._distortion

    @property
    def channel_mix(self) -> ChannelMix:
        """Property which returns the :class:`~wavelink.ChannelMix` filter associated with this Filters payload."""
        return self._channel_mix

    @property
    def low_pass(self) -> LowPass:
        """Property which returns the :class:`~wavelink.LowPass` filter associated with this Filters payload."""
        return self._low_pass

    @property
    def plugin_filters(self) -> PluginFilters:
        """Property which returns the :class:`~wavelink.PluginFilters` filters associated with this Filters payload."""
        return self._plugin_filters

    def __call__(self) -> FilterPayload:
        payload: FilterPayload = {
            "volume": self._volume,
            "equalizer": list(self._equalizer._payload.values()),
            "karaoke": self._karaoke._payload,
            "timescale": self._timescale._payload,
            "tremolo": self._tremolo._payload,
            "vibrato": self._vibrato._payload,
            "rotation": self._rotation._payload,
            "distortion": self._distortion._payload,
            "channelMix": self._channel_mix._payload,
            "lowPass": self._low_pass._payload,
            "pluginFilters": self._plugin_filters._payload,
        }

        for key, value in payload.copy().items():
            if not value:
                del payload[key]

        return payload

    def __repr__(self) -> str:
        return (
            f"<Filters: volume={self._volume}, equalizer={self._equalizer!r}, karaoke={self._karaoke!r},"
            f" timescale={self._timescale!r}, tremolo={self._tremolo!r}, vibrato={self._vibrato!r},"
            f" rotation={self._rotation!r}, distortion={self._distortion!r}, channel_mix={self._channel_mix!r},"
            f" low_pass={self._low_pass!r}, plugin_filters={self._plugin_filters!r}>"
        )



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import enum


__all__ = ("NodeStatus", "TrackSource", "DiscordVoiceCloseType", "AutoPlayMode", "QueueMode")


class NodeStatus(enum.Enum):
    """Enum representing the connection status of a Node.

    Attributes
    ----------
    DISCONNECTED
        The Node has been disconnected or has never been connected previously.
    CONNECTING
        The Node is currently attempting to connect.
    CONNECTED
        The Node is currently connected.
    """

    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2


class TrackSource(enum.Enum):
    """Enum representing a :class:`Playable` source.

    Attributes
    ----------
    YouTube
        A source representing a track that comes from YouTube.
    YouTubeMusic
        A source representing a track that comes from YouTube Music.
    SoundCloud
        A source representing a track that comes from SoundCloud.
    """

    YouTube = 0
    YouTubeMusic = 1
    SoundCloud = 2


class DiscordVoiceCloseType(enum.Enum):
    """Enum representing the various Discord Voice Websocket Close Codes.

    Attributes
    ----------
    CLOSE_NORMAL
        1000
    UNKNOWN_OPCODE
        4001
    FAILED_DECODE_PAYLOAD
        4002
    NOT_AUTHENTICATED
        4003
    AUTHENTICATION_FAILED
        4004
    ALREADY_AUTHENTICATED
        4005
    SESSION_INVALID
        4006
    SESSION_TIMEOUT
        4009
    SERVER_NOT_FOUND
        4011
    UNKNOWN_PROTOCOL
        4012
    DISCONNECTED
        4014
    VOICE_SERVER_CRASHED
        4015
    UNKNOWN_ENCRYPTION_MODE
        4016
    """

    CLOSE_NORMAL = 1000  # Not Discord but standard websocket
    UNKNOWN_OPCODE = 4001
    FAILED_DECODE_PAYLOAD = 4002
    NOT_AUTHENTICATED = 4003
    AUTHENTICATION_FAILED = 4004
    ALREADY_AUTHENTICATED = 4005
    SESSION_INVALID = 4006
    SESSION_TIMEOUT = 4009
    SERVER_NOT_FOUND = 4011
    UNKNOWN_PROTOCOL = 4012
    DISCONNECTED = 4014
    VOICE_SERVER_CRASHED = 4015
    UNKNOWN_ENCRYPTION_MODE = 4016


class AutoPlayMode(enum.Enum):
    """Enum representing the various AutoPlay modes.

    Attributes
    ----------
    enabled
        When enabled, AutoPlay will work fully autonomously and fill the auto_queue with recommended tracks.
        If a song is put into a players standard queue, AutoPlay will use it as a priority.
    partial
        When partial, AutoPlay will work fully autonomously but **will not** fill the auto_queue with
        recommended tracks.
    disabled
        When disabled, AutoPlay will not do anything automatically.
    """

    enabled = 0
    partial = 1
    disabled = 2


class QueueMode(enum.Enum):
    """Enum representing the various modes on :class:`wavelink.Queue`

    Attributes
    ----------
    normal
        When set, the queue will not loop either track or history. This is the default.
    loop
        When set, the track will continuously loop.
    loop_all
        When set, the queue will continuously loop through all tracks.
    """

    normal = 0
    loop = 1
    loop_all = 2



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeAlias, overload

import yarl

import wavelink

from .enums import TrackSource
from .utils import ExtrasNamespace


if TYPE_CHECKING:
    from collections.abc import Iterator

    from .node import Node
    from .types.tracks import (
        PlaylistInfoPayload,
        PlaylistPayload,
        TrackInfoPayload,
        TrackPayload,
    )


__all__ = ("Search", "Album", "Artist", "Playable", "Playlist", "PlaylistInfo")


_source_mapping: dict[TrackSource | str | None, str] = {
    TrackSource.YouTube: "ytsearch",
    TrackSource.SoundCloud: "scsearch",
    TrackSource.YouTubeMusic: "ytmsearch",
}


Search: TypeAlias = "list[Playable] | Playlist"


class Album:
    """Container class representing Album data received via Lavalink.

    Attributes
    ----------
    name: str | None
        The album name. Could be ``None``.
    url: str | None
        The album url. Could be ``None``.
    """

    def __init__(self, *, data: dict[Any, Any]) -> None:
        self.name: str | None = data.get("albumName")
        self.url: str | None = data.get("albumUrl")


class Artist:
    """Container class representing Artist data received via Lavalink.

    Attributes
    ----------
    url: str | None
        The artist url. Could be ``None``.
    artwork: str | None
        The artist artwork url. Could be ``None``.
    """

    def __init__(self, *, data: dict[Any, Any]) -> None:
        self.url: str | None = data.get("artistUrl")
        self.artwork: str | None = data.get("artistArtworkUrl")


class Playable:
    """The Wavelink Playable object which represents all tracks in Wavelink 3.

    .. note::

        You should not construct this class manually.

    .. container:: operations

        .. describe:: str(track)

            The title of this playable.

        .. describe:: repr(track)

            The official string representation of this playable.

        .. describe:: track == other

            Whether this track is equal to another. Checks both the track encoding and identifier.
    """

    def __init__(self, data: TrackPayload, *, playlist: PlaylistInfo | None = None) -> None:
        info: TrackInfoPayload = data["info"]

        self._encoded: str = data["encoded"]
        self._identifier: str = info["identifier"]
        self._is_seekable: bool = info["isSeekable"]
        self._author: str = info["author"]
        self._length: int = info["length"]
        self._is_stream: bool = info["isStream"]
        self._position: int = info["position"]
        self._title: str = info["title"]
        self._uri: str | None = info.get("uri")
        self._artwork: str | None = info.get("artworkUrl")
        self._isrc: str | None = info.get("isrc")
        self._source: str = info["sourceName"]

        plugin: dict[Any, Any] = data["pluginInfo"]
        self._album: Album = Album(data=plugin)
        self._artist: Artist = Artist(data=plugin)

        self._preview_url: str | None = plugin.get("previewUrl")
        self._is_preview: bool | None = plugin.get("isPreview")

        self._playlist = playlist
        self._recommended: bool = False

        self._extras: ExtrasNamespace = ExtrasNamespace(data.get("userData", {}))

        self._raw_data = data

    def __hash__(self) -> int:
        return hash(self.encoded)

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return f"Playable(source={self.source}, title={self.title}, identifier={self.identifier})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Playable):
            return NotImplemented

        return self.encoded == other.encoded or self.identifier == other.identifier

    @property
    def encoded(self) -> str:
        """Property returning the encoded track string from Lavalink."""
        return self._encoded

    @property
    def identifier(self) -> str:
        """Property returning the identifier of this track from its source.

        E.g. YouTube ID or Spotify ID.
        """
        return self._identifier

    @property
    def is_seekable(self) -> bool:
        """Property returning a bool whether this track can be used in seeking."""
        return self._is_seekable

    @property
    def author(self) -> str:
        """Property returning the name of the author of this track."""
        return self._author

    @property
    def length(self) -> int:
        """Property returning the tracks duration in milliseconds as an int."""
        return self._length

    @property
    def is_stream(self) -> bool:
        """Property returning a bool indicating whether this track is a stream."""
        return self._is_stream

    @property
    def position(self) -> int:
        """Property returning starting position of this track in milliseconds as an int."""
        return self._position

    @property
    def title(self) -> str:
        """Property returning the title/name of this track."""
        return self._title

    @property
    def uri(self) -> str | None:
        """Property returning the URL to this track. Could be ``None``."""
        return self._uri

    @property
    def artwork(self) -> str | None:
        """Property returning the URL of the artwork of this track. Could be ``None``."""
        return self._artwork

    @property
    def isrc(self) -> str | None:
        """Property returning the ISRC (International Standard Recording Code) of this track. Could be ``None``."""
        return self._isrc

    @property
    def source(self) -> str:
        """Property returning the source of this track as a ``str``.

        E.g. "spotify" or "youtube".
        """
        return self._source

    @property
    def album(self) -> Album:
        """Property returning album data for this track."""
        return self._album

    @property
    def artist(self) -> Artist:
        """Property returning artist data for this track."""
        return self._artist

    @property
    def preview_url(self) -> str | None:
        """Property returning the preview URL for this track. Could be ``None``."""
        return self._preview_url

    @property
    def is_preview(self) -> bool | None:
        """Property returning a bool indicating if this track is a preview. Could be ``None`` if unknown."""
        return self._is_preview

    @property
    def playlist(self) -> PlaylistInfo | None:
        """Property returning a :class:`wavelink.PlaylistInfo`. Could be ``None``
        if this track is not a part of a playlist.
        """
        return self._playlist

    @property
    def recommended(self) -> bool:
        """Property returning a bool indicating whether this track was recommended via AutoPlay."""
        return self._recommended

    @property
    def extras(self) -> ExtrasNamespace:
        """Property returning a :class:`~wavelink.ExtrasNamespace` of extras for this :class:`Playable`.

        You can set this property with a :class:`dict` of valid :class:`str` keys to any valid ``JSON`` value,
        or a :class:`~wavelink.ExtrasNamespace`.

        If a dict is passed, it will be converted into an :class:`~wavelink.ExtrasNamespace`,
        which can be converted back to a dict with dict(...). Additionally, you can also use list or tuple on
        :class:`~wavelink.ExtrasNamespace`.

        The extras dict will be sent to Lavalink as the ``userData`` field.


        .. warning::

            This is only available when using Lavalink 4+ (**Non BETA**) versions.


        Examples
        --------

            .. code:: python

                track: wavelink.Playable = wavelink.Playable.search("QUERY")
                track.extras = {"requester_id": 1234567890}

                # later...
                print(track.extras.requester_id)
                # or
                print(dict(track.extras)["requester_id"])


        .. versionadded:: 3.1.0
        """
        return self._extras

    @extras.setter
    def extras(self, __value: ExtrasNamespace | dict[str, Any]) -> None:
        if isinstance(__value, ExtrasNamespace):
            self._extras = __value
        else:
            self._extras = ExtrasNamespace(__value)

    @property
    def raw_data(self) -> TrackPayload:
        """The raw data for this ``Playable`` received via ``Lavalink``.

        You can use this data to reconstruct this ``Playable`` object.


        Examples
        --------

            .. code:: python3

                # For example purposes...
                old_data = track.raw_data

                # Later...
                track: wavelink.Playable = wavelink.Playable(old_data)


        .. versionadded:: 3.2.0
        """
        return self._raw_data

    @classmethod
    async def search(
        cls, query: str, /, *, source: TrackSource | str | None = TrackSource.YouTubeMusic, node: Node | None = None
    ) -> Search:
        """Search for a list of :class:`~wavelink.Playable` or a :class:`~wavelink.Playlist`, with the given query.

        .. note::

            This method differs from :meth:`wavelink.Pool.fetch_tracks` in that it will apply a relevant search prefix
            for you when a URL is **not** provided. This prefix can be controlled via the ``source`` keyword argument.


        .. note::

            This method of searching is preferred over, :meth:`wavelink.Pool.fetch_tracks`.


        Parameters
        ----------
        query: str
            The query to search tracks for. If this is **not** a URL based search this method will provide an
            appropriate search prefix based on what is provided to the ``source`` keyword only parameter,
            or it's default.

            If this query **is a URL**, a search prefix will **not** be used.
        source: :class:`TrackSource` | str | None
            This parameter determines which search prefix to use when searching for tracks.
            If ``None`` is provided, no prefix will be used, however this behaviour is default regardless of what
            is provided **when a URL is found**.

            For basic searches, E.g. YouTube, YouTubeMusic and SoundCloud, see: :class:`wavelink.TrackSource`.
            Otherwise, a ``str`` may be provided for plugin based searches, E.g. "spsearch:" for the
            LavaSrc Spotify based search.

            Defaults to :attr:`wavelink.TrackSource.YouTubeMusic` which is equivalent to "ytmsearch:".
        node: :class:`~wavelink.Node` | None
            An optional :class:`~wavelink.Node` to use when searching for tracks. Defaults to ``None``, which uses
            the :class:`~wavelink.Pool`'s automatic node selection.


        Returns
        -------
        :class:`wavelink.Search`
            A union of either list[:class:`Playable`] or :class:`Playlist`. Could return and empty list,
            if no tracks or playlist were found.

        Raises
        ------
        LavalinkLoadException
            Exception raised when Lavalink fails to load results based on your query.


        Examples
        --------

        .. code:: python3

            # Search for tracks, with the default "ytsearch:" prefix.
            tracks: wavelink.Search = await wavelink.Playable.search("Ocean Drive")
            if not tracks:
                # No tracks were found...
                ...

            # Search for tracks, with a URL.
            tracks: wavelink.Search = await wavelink.Playable.search("https://www.youtube.com/watch?v=KDxJlW6cxRk")
            ...

            # Search for tracks, using Spotify and the LavaSrc Plugin.
            tracks: wavelink.Search = await wavelink.Playable.search("4b93D55xv3YCH5mT4p6HPn", source="spsearch")
            ...

            # Search for tracks, using Spotify and the LavaSrc Plugin, with a URL.
            tracks: wavelink.Search = await wavelink.Playable.search("https://open.spotify.com/track/4b93D55xv3YCH5mT4p6HPn")
            ...

            # Search for a playlist, using Spotify and the LavaSrc Plugin.
            # or alternatively any other playlist URL from another source like YouTube.
            tracks: wavelink.Search = await wavelink.Playable.search("https://open.spotify.com/playlist/37i9dQZF1DWXRqgorJj26U")
            ...


        .. versionchanged:: 3.0.0

            This method has been changed significantly in version ``3.0.0``. This method has been simplified to provide
            an easier interface for searching tracks. See the above documentation and examples.

            You can no longer provide a :class:`wavelink.Node` to use for searching as this method will now select the
            most appropriate node from the :class:`wavelink.Pool`.
        """
        prefix: TrackSource | str | None = _source_mapping.get(source, source)
        check = yarl.URL(query)

        if check.host:
            tracks: Search = await wavelink.Pool.fetch_tracks(query, node=node)
            return tracks

        if not prefix:
            term: str = query
        else:
            assert not isinstance(prefix, TrackSource)
            term: str = f"{prefix.removesuffix(':')}:{query}"

        tracks: Search = await wavelink.Pool.fetch_tracks(term, node=node)
        return tracks


class Playlist:
    """The wavelink Playlist container class.

    This class is created and returned via both :meth:`Playable.search` and :meth:`wavelink.Pool.fetch_tracks`.

    It contains various information about the playlist and a list of :class:`Playable` that can be used directly in
    :meth:`wavelink.Player.play`. See below for various supported operations.


    .. warning::

        You should not instantiate this class manually,
        use :meth:`Playable.search` or :meth:`wavelink.Pool.fetch_tracks` instead.


    .. warning::

        You can not use ``.search`` directly on this class, see: :meth:`Playable.search`.


    .. note::

        This class can be directly added to :class:`wavelink.Queue` identical to :class:`Playable`. When added,
        all tracks contained in this playlist, will be individually added to the :class:`wavelink.Queue`.


    .. container:: operations

        .. describe:: str(x)

            Return the name associated with this playlist.

        .. describe:: repr(x)

            Return the official string representation of this playlist.

        .. describe:: x == y

            Compare the equality of playlist.

        .. describe:: len(x)

            Return an integer representing the amount of tracks contained in this playlist.

        .. describe:: x[0]

            Return a track contained in this playlist with the given index.

        .. describe:: x[0:2]

            Return a slice of tracks contained in this playlist.

        .. describe:: for x in y

            Iterate over the tracks contained in this playlist.

        .. describe:: reversed(x)

            Reverse the tracks contained in this playlist.

        .. describe:: x in y

            Check if a :class:`Playable` is contained in this playlist.


    Attributes
    ----------
    name: str
        The name of this playlist.
    selected: int
        The index of the selected track from Lavalink.
    tracks: list[:class:`Playable`]
        A list of :class:`Playable` contained in this playlist.
    type: str | None
        An optional ``str`` identifying the type of playlist this is. Only available when a plugin is used.
    url: str | None
        An optional ``str`` to the URL of this playlist. Only available when a plugin is used.
    artwork: str | None
        An optional ``str`` to the artwork of this playlist. Only available when a plugin is used.
    author: str | None
        An optional ``str`` of the author of this playlist. Only available when a plugin is used.
    """

    def __init__(self, data: PlaylistPayload) -> None:
        info: PlaylistInfoPayload = data["info"]
        self.name: str = info["name"]
        self.selected: int = info["selectedTrack"]

        playlist_info: PlaylistInfo = PlaylistInfo(data)
        self.tracks: list[Playable] = [Playable(data=track, playlist=playlist_info) for track in data["tracks"]]

        plugin: dict[Any, Any] = data["pluginInfo"]
        self.type: str | None = plugin.get("type")
        self.url: str | None = plugin.get("url")
        self.artwork: str | None = plugin.get("artworkUrl")
        self.author: str | None = plugin.get("author")

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Playlist(name={self.name}, tracks={len(self.tracks)})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Playlist):
            return NotImplemented

        return self.name == other.name and self.tracks == other.tracks

    def __len__(self) -> int:
        return len(self.tracks)

    @overload
    def __getitem__(self, index: int) -> Playable: ...

    @overload
    def __getitem__(self, index: slice) -> list[Playable]: ...

    def __getitem__(self, index: int | slice) -> Playable | list[Playable]:
        return self.tracks[index]

    def __iter__(self) -> Iterator[Playable]:
        return self.tracks.__iter__()

    def __reversed__(self) -> Iterator[Playable]:
        return self.tracks.__reversed__()

    def __contains__(self, item: Playable) -> bool:
        return item in self.tracks

    def pop(self, index: int = -1) -> Playable:
        return self.tracks.pop(index)

    def track_extras(self, **attrs: object) -> None:
        """Method which sets attributes to all :class:`Playable` in this playlist, with the provided keyword arguments.

        This is useful when you need to attach state to your :class:`Playable`, E.g. create a requester attribute.

        .. warning::

            If you try to override any existing property of :class:`Playable` this method will fail.


        Parameters
        ----------
        **attrs
            The keyword arguments to set as attribute name=value on each :class:`Playable`.

        Examples
        --------

            .. code:: python3

                playlist.track_extras(requester=ctx.author)

                track: wavelink.Playable = playlist[0]
                print(track.requester)
        """
        for track in self.tracks:
            for name, value in attrs.items():
                setattr(track, name, value)

    @property
    def extras(self) -> ExtrasNamespace:
        """Property returning a :class:`~wavelink.ExtrasNamespace` of extras for this :class:`Playlist`.

        You can set this property with a :class:`dict` of valid :class:`str` keys to any valid ``JSON`` value,
        or a :class:`~wavelink.ExtrasNamespace`.

        If a dict is passed, it will be converted into an :class:`~wavelink.ExtrasNamespace`,
        which can be converted back to a dict with ``dict(...)``. Additionally, you can also use list or tuple on
        :class:`~wavelink.ExtrasNamespace`.

        The extras dict will be sent to Lavalink as the ``userData`` field for each track in the playlist.


        .. warning::

            This is only available when using Lavalink 4+ (**Non BETA**) versions.


        Examples
        --------

            .. code:: python

                playlist: wavelink.Search = wavelink.Playable.search("QUERY")
                playlist.extras = {"requester_id": 1234567890}

                # later...
                print(track.extras.requester_id)
                # or
                print(dict(track.extras)["requester_id"])


        .. versionadded:: 3.2.0
        """
        return self._extras

    @extras.setter
    def extras(self, __value: ExtrasNamespace | dict[str, Any]) -> None:
        if isinstance(__value, ExtrasNamespace):
            self._extras = __value
        else:
            self._extras = ExtrasNamespace(__value)

        for track in self.tracks:
            track.extras = __value


class PlaylistInfo:
    """The wavelink PlaylistInfo container class.

    It contains various information about the playlist but **does not** contain the tracks associated with this
    playlist.

    This class is used to provided information about the original :class:`wavelink.Playlist` on tracks.

    Attributes
    ----------
    name: str
        The name of this playlist.
    selected: int
        The index of the selected track from Lavalink.
    tracks: int
        The amount of tracks this playlist originally contained.
    type: str | None
        An optional ``str`` identifying the type of playlist this is. Only available when a plugin is used.
    url: str | None
        An optional ``str`` to the URL of this playlist. Only available when a plugin is used.
    artwork: str | None
        An optional ``str`` to the artwork of this playlist. Only available when a plugin is used.
    author: str | None
        An optional ``str`` of the author of this playlist. Only available when a plugin is used.
    """

    __slots__ = ("name", "selected", "tracks", "type", "url", "artwork", "author")

    def __init__(self, data: PlaylistPayload) -> None:
        info: PlaylistInfoPayload = data["info"]
        self.name: str = info["name"]
        self.selected: int = info["selectedTrack"]

        self.tracks: int = len(data["tracks"])

        plugin: dict[Any, Any] = data["pluginInfo"]
        self.type: str | None = plugin.get("type")
        self.url: str | None = plugin.get("url")
        self.artwork: str | None = plugin.get("artworkUrl")
        self.author: str | None = plugin.get("author")

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"PlaylistInfo(name={self.name}, tracks={self.tracks})"

    def __len__(self) -> int:
        return self.tracks



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .types.response import ErrorResponse, LoadedErrorPayload


__all__ = (
    "WavelinkException",
    "NodeException",
    "InvalidClientException",
    "AuthorizationFailedException",
    "InvalidNodeException",
    "LavalinkException",
    "LavalinkLoadException",
    "InvalidChannelStateException",
    "ChannelTimeoutException",
    "QueueEmpty",
)


class WavelinkException(Exception):
    """Base wavelink Exception class.

    All wavelink exceptions derive from this exception.
    """


class NodeException(WavelinkException):
    """Error raised when an Unknown or Generic error occurs on a Node.

    This exception may be raised when an error occurs reaching your Node.

    Attributes
    ----------
    status: int | None
        The status code received when making a request. Could be None.
    """

    def __init__(self, msg: str | None = None, status: int | None = None) -> None:
        super().__init__(msg)

        self.status = status


class InvalidClientException(WavelinkException):
    """Exception raised when an invalid :class:`discord.Client`
    is provided while connecting a :class:`wavelink.Node`.
    """


class AuthorizationFailedException(WavelinkException):
    """Exception raised when Lavalink fails to authenticate a :class:`~wavelink.Node`, with the provided password."""


class InvalidNodeException(WavelinkException):
    """Exception raised when a :class:`Node` is tried to be retrieved from the
    :class:`Pool` without existing, or the ``Pool`` is empty.
    """


class LavalinkException(WavelinkException):
    """Exception raised when Lavalink returns an invalid response.

    Attributes
    ----------
    status: int
        The response status code.
    reason: str | None
        The response reason. Could be ``None`` if no reason was provided.
    """

    def __init__(self, msg: str | None = None, /, *, data: ErrorResponse) -> None:
        self.timestamp: int = data["timestamp"]
        self.status: int = data["status"]
        self.error: str = data["error"]
        self.trace: str | None = data.get("trace")
        self.path: str = data["path"]

        if not msg:
            msg = f"Failed to fulfill request to Lavalink: status={self.status}, reason={self.error}, path={self.path}"

        super().__init__(msg)


class LavalinkLoadException(WavelinkException):
    """Exception raised when an error occurred loading tracks via Lavalink.

    Attributes
    ----------
    error: str
        The error message from Lavalink.
    severity: str
        The severity of this error sent via Lavalink.
    cause: str
        The cause of this error sent via Lavalink.
    """

    def __init__(self, msg: str | None = None, /, *, data: LoadedErrorPayload) -> None:
        self.error: str = data["message"]
        self.severity: str = data["severity"]
        self.cause: str = data["cause"]

        if not msg:
            msg = f"Failed to Load Tracks: error={self.error}, severity={self.severity}, cause={self.cause}"

        super().__init__(msg)


class InvalidChannelStateException(WavelinkException):
    """Exception raised when a :class:`~wavelink.Player` tries to connect to an invalid channel or
    has invalid permissions to use this channel.
    """


class ChannelTimeoutException(WavelinkException):
    """Exception raised when connecting to a voice channel times out."""


class QueueEmpty(WavelinkException):
    """Exception raised when you try to retrieve from an empty queue."""
